@@ -1,4 +1,5 @@
 """Decode JSON data."""
+
 # Copyright 2019, B-Open Solutions srl.
 # Copyright 2022, European Union.
 #
@@ -39,6 +40,7 @@ class DecodeError(Exception):
 def decode_python_object(obj: Dict[str, Any]) -> Any:
     if obj.get("type") == "python_object" and "fully_qualified_name" in obj:
         return import_object(obj["fully_qualified_name"])
+    return None
 
 
 def decode_python_call(obj: Dict[str, Any]) -> Any:
@@ -50,6 +52,7 @@ def decode_python_call(obj: Dict[str, Any]) -> Any:
         args = obj.get("args", ())
         kwargs = obj.get("kwargs", {})
         return func(*args, **kwargs)
+    return None
 
 
 FILECACHE_DECODERS: List[Callable[[Dict[str, Any]], Any]] = [
@@ -60,14 +63,13 @@ FILECACHE_DECODERS: List[Callable[[Dict[str, Any]], Any]] = [
 
 def object_hook(obj: Dict[str, Any]) -> Any:
     """Decode deserialized JSON data (``dict``)."""
-    if isinstance(obj, dict):
-        for decoder in reversed(FILECACHE_DECODERS):
-            try:
-                result = decoder(obj)
-                if result is not None:
-                    return result
-            except Exception as ex:
-                raise DecodeError(ex)
+    for decoder in reversed(FILECACHE_DECODERS):
+        try:
+            result = decoder(obj)
+            if result is not None:
+                return result
+        except Exception as ex:
+            raise DecodeError(ex) from ex
 
     return obj
 
