@@ -1,4 +1,5 @@
 """Additional encoders that need optional dependencies."""
+
 # Copyright 2019, B-Open Solutions srl.
 # Copyright 2022, European Union.
 #
@@ -14,18 +15,18 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-
 import functools
 import inspect
 import io
 import mimetypes
-import os
+import pathlib
 import posixpath
 import tempfile
 from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union, cast
 
 import fsspec
 import fsspec.implementations.local
+import pydantic
 
 from . import config, encode, utils
 
@@ -47,17 +48,23 @@ except ImportError:
 F = TypeVar("F", bound=Callable[..., Any])
 
 _UNION_IO_TYPES = Union[
-    io.BufferedReader,
-    io.TextIOWrapper,
+    io.RawIOBase,
+    io.BufferedIOBase,
+    io.TextIOBase,
     fsspec.spec.AbstractBufferedFile,
     fsspec.implementations.local.LocalFileOpener,
 ]
 
-# Add netcdf, grib, and zarr to mimetypes
-mimetypes.add_type("application/netcdf", ".nc", strict=True)
-for ext in (".grib", ".grb", ".grb1", ".grb2"):
-    mimetypes.add_type("application/x-grib", ext, strict=False)
-mimetypes.add_type("application/vnd+zarr", ".zarr", strict=False)
+
+def _add_ext_to_mimetypes() -> None:
+    """Add netcdf, grib, and zarr to mimetypes."""
+    mimetypes.add_type("application/netcdf", ".nc", strict=True)
+    for ext in (".grib", ".grb", ".grb1", ".grb2"):
+        mimetypes.add_type("application/x-grib", ext, strict=False)
+    mimetypes.add_type("application/vnd+zarr", ".zarr", strict=False)
+
+
+_add_ext_to_mimetypes()
 
 
 def _requires_xarray_and_dask(func: F) -> F:
@@ -72,6 +79,14 @@ def _requires_xarray_and_dask(func: F) -> F:
     return cast(F, wrapper)
 
 
+class FileInfoModel(pydantic.BaseModel):
+    type: str
+    href: str
+    file_checksum: int = pydantic.Field(..., alias="file:checksum")
+    file_size: int = pydantic.Field(..., alias="file:size")
+    file_local_path: str = pydantic.Field(..., alias="file:local_path")
+
+
 def _dictify_file(fs: fsspec.AbstractFileSystem, local_path: str) -> Dict[str, Any]:
 
     filetype = mimetypes.guess_type(local_path, strict=False)[0]
@@ -85,8 +100,8 @@ def _dictify_file(fs: fsspec.AbstractFileSystem, local_path: str) -> Dict[str, A
     file_dict = {
         "type": filetype,
         "href": posixpath.join(
-            config.SETTINGS["cache_files_urlpath_readonly"]
-            or config.SETTINGS["cache_files_urlpath"],
+            config.SETTINGS.get().cache_files_urlpath_readonly
+            or config.SETTINGS.get().cache_files_urlpath,
             posixpath.basename(local_path),
         ),
         "file:checksum": fs.checksum(local_path),
@@ -94,25 +109,7 @@ def _dictify_file(fs: fsspec.AbstractFileSystem, local_path: str) -> Dict[str, A
         "file:local_path": local_path,
     }
 
-    return file_dict
-
-
-def _are_file_args(*args: Any) -> bool:
-    try:
-        return (
-            len(args) <= 2
-            and all(isinstance(arg, dict) for arg in args)
-            and set(args[0])
-            == {
-                "type",
-                "href",
-                "file:checksum",
-                "file:size",
-                "file:local_path",
-            }
-        )
-    except TypeError:
-        return False
+    return FileInfoModel(**file_dict).dict(by_alias=True)
 
 
 def _get_fs_and_urlpath(
@@ -123,7 +120,7 @@ def _get_fs_and_urlpath(
 
     urlpath = file_json["file:local_path"]
     if storage_options is None:
-        storage_options = config.SETTINGS["cache_files_storage_options"]
+        storage_options = config.SETTINGS.get().cache_files_storage_options
 
     if not validate:
         fs, _, _ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
@@ -187,36 +184,39 @@ def decode_io_object(
 
 
 @_requires_xarray_and_dask
-def _store_xr_dataset(
+def _maybe_store_xr_dataset(
     obj: "xr.Dataset", fs: fsspec.AbstractFileSystem, urlpath: str, filetype: str
 ) -> None:
-    if filetype == "application/vnd+zarr":
-        # Write directly on any filesystem
-        mapper = fs.get_mapper(urlpath)
-        obj.to_zarr(mapper, consolidated=True)
-    else:
-        # Need a tmp local copy to write on a different filesystem
-        with tempfile.TemporaryDirectory() as tmpdirname:
-            basename = os.path.basename(urlpath)
-            tmpfilename = os.path.join(tmpdirname, basename)
-
-            if filetype == "application/netcdf":
-                obj.to_netcdf(tmpfilename)
-            elif filetype == "application/x-grib":
-                import cfgrib.xarray_to_grib
-
-                cfgrib.xarray_to_grib.to_grib(obj, tmpfilename)
+    with utils._Locker(fs, urlpath) as file_exists:
+        if not file_exists:
+            if filetype == "application/vnd+zarr":
+                # Write directly on any filesystem
+                mapper = fs.get_mapper(urlpath)
+                obj.to_zarr(mapper, consolidated=True)
             else:
-                # Should never get here! xarray_cache_type is checked in config.py
-                raise ValueError(f"type {filetype!r} is NOT supported.")
+                # Need a tmp local copy to write on a different filesystem
+                with tempfile.TemporaryDirectory() as tmpdirname:
+                    tmpfilename = str(
+                        pathlib.Path(tmpdirname) / pathlib.Path(urlpath).name
+                    )
 
-            if fs == fsspec.filesystem("file"):
-                fsspec.filesystem("file").mv(tmpfilename, urlpath)
-            else:
-                with fsspec.open(tmpfilename, "rb") as f_in, fs.open(
-                    urlpath, "wb"
-                ) as f_out:
-                    utils.copy_buffered_file(f_in, f_out)
+                    if filetype == "application/netcdf":
+                        obj.to_netcdf(tmpfilename)
+                    elif filetype == "application/x-grib":
+                        import cfgrib.xarray_to_grib
+
+                        cfgrib.xarray_to_grib.to_grib(obj, tmpfilename)
+                    else:
+                        # Should never get here! xarray_cache_type is checked in config.py
+                        raise ValueError(f"type {filetype!r} is NOT supported.")
+
+                    if fs == fsspec.filesystem("file"):
+                        fsspec.filesystem("file").mv(tmpfilename, urlpath)
+                    else:
+                        with fsspec.open(tmpfilename, "rb") as f_in, fs.open(
+                            urlpath, "wb"
+                        ) as f_out:
+                            utils.copy_buffered_file(f_in, f_out)
 
 
 @_requires_xarray_and_dask
@@ -225,15 +225,17 @@ def dictify_xr_dataset(obj: "xr.Dataset") -> Dict[str, Any]:
     with dask.config.set({"tokenize.ensure-deterministic": True}):
         root = dask.base.tokenize(obj)  # type: ignore[no-untyped-call]
 
-    filetype = config.SETTINGS["xarray_cache_type"]
+    filetype = config.SETTINGS.get().xarray_cache_type
     ext = mimetypes.guess_extension(filetype, strict=False)
-    urlpath_out = posixpath.join(config.SETTINGS["cache_files_urlpath"], f"{root}{ext}")
+    urlpath_out = posixpath.join(
+        config.SETTINGS.get().cache_files_urlpath, f"{root}{ext}"
+    )
 
     fs_out, _, _ = fsspec.get_fs_token_paths(
-        urlpath_out, storage_options=config.SETTINGS["cache_files_storage_options"]
+        urlpath_out,
+        storage_options=config.SETTINGS.get().cache_files_storage_options,
     )
-    if not fs_out.exists(urlpath_out):
-        _store_xr_dataset(obj, fs_out, urlpath_out, filetype)
+    _maybe_store_xr_dataset(obj, fs_out, urlpath_out, filetype)
 
     file_json = _dictify_file(fs_out, urlpath_out)
     if filetype == "application/vnd+zarr":
@@ -248,53 +250,64 @@ def dictify_xr_dataset(obj: "xr.Dataset") -> Dict[str, Any]:
     return encode.dictify_python_call(
         decode_xr_dataset,
         file_json,
-        storage_options=config.SETTINGS["cache_files_storage_options"],
+        storage_options=config.SETTINGS.get().cache_files_storage_options,
         **kwargs,
     )
 
 
-def _store_io_object(
+def _maybe_store_file_object(
     fs_in: fsspec.AbstractFileSystem,
     urlpath_in: str,
     fs_out: fsspec.AbstractFileSystem,
     urlpath_out: str,
 ) -> None:
-    if fs_in == fs_out:
-        if config.SETTINGS["io_delete_original"]:
-            fs_in.mv(urlpath_in, urlpath_out)
-        else:
-            fs_in.cp(urlpath_in, urlpath_out)
-    else:
-        with fs_in.open(urlpath_in, "rb") as f_in, fs_out.open(
-            urlpath_out, "wb"
-        ) as f_out:
+    with utils._Locker(fs_out, urlpath_out) as file_exists:
+        if not file_exists:
+            if fs_in == fs_out:
+                if config.SETTINGS.get().io_delete_original:
+                    fs_in.mv(urlpath_in, urlpath_out)
+                else:
+                    fs_in.cp(urlpath_in, urlpath_out)
+            else:
+                with fs_in.open(urlpath_in, "rb") as f_in, fs_out.open(
+                    urlpath_out, "wb"
+                ) as f_out:
+                    utils.copy_buffered_file(f_in, f_out)
+                if config.SETTINGS.get().io_delete_original and fs_in.exists(
+                    urlpath_in
+                ):
+                    fs_in.rm(urlpath_in)
+
+
+def _maybe_store_io_object(
+    f_in: _UNION_IO_TYPES,
+    fs_out: fsspec.AbstractFileSystem,
+    urlpath_out: str,
+) -> None:
+    with utils._Locker(fs_out, urlpath_out) as file_exists:
+        if not file_exists:
+            f_out = fs_out.open(urlpath_out, "wb")
             utils.copy_buffered_file(f_in, f_out)
-        if config.SETTINGS["io_delete_original"]:
-            fs_in.rm(urlpath_in)
 
 
 def dictify_io_object(obj: _UNION_IO_TYPES) -> Dict[str, Any]:
     """Encode a file object to JSON deserialized data (``dict``)."""
-    if "w" in obj.mode:
-        raise ValueError("write-mode objects can NOT be cached.")
-
-    if isinstance(obj, (io.BufferedReader, io.TextIOWrapper)):
-        urlpath_in = obj.name
-        fs_in = fsspec.filesystem("file")
-    else:
-        urlpath_in = obj.path
-        fs_in = obj.fs
-    urlpath_in = str(urlpath_in)
-
-    root = fs_in.checksum(urlpath_in)
-    _, ext = os.path.splitext(urlpath_in)
-    urlpath_out = posixpath.join(config.SETTINGS["cache_files_urlpath"], f"{root}{ext}")
-
+    cache_files_urlpath = config.SETTINGS.get().cache_files_urlpath
     fs_out, _, _ = fsspec.get_fs_token_paths(
-        urlpath_out, storage_options=config.SETTINGS["cache_files_storage_options"]
+        cache_files_urlpath,
+        storage_options=config.SETTINGS.get().cache_files_storage_options,
     )
-    if not fs_out.exists(urlpath_out):
-        _store_io_object(fs_in, urlpath_in, fs_out, urlpath_out)
+
+    if hasattr(obj, "path") or hasattr(obj, "name"):
+        urlpath_in = obj.path if hasattr(obj, "path") else obj.name  # type: ignore[union-attr]
+        fs_in = getattr(obj, "fs", fsspec.filesystem("file"))
+        root = fs_in.checksum(urlpath_in)
+        ext = pathlib.Path(urlpath_in).suffix
+        urlpath_out = posixpath.join(cache_files_urlpath, f"{root}{ext}")
+        _maybe_store_file_object(fs_in, urlpath_in, fs_out, urlpath_out)
+    else:
+        urlpath_out = posixpath.join(cache_files_urlpath, f"{hash(obj)}")
+        _maybe_store_io_object(obj, fs_out, urlpath_out)
 
     file_json = _dictify_file(fs_out, urlpath_out)
     params = inspect.signature(open).parameters
@@ -303,7 +316,7 @@ def dictify_io_object(obj: _UNION_IO_TYPES) -> Dict[str, Any]:
     return encode.dictify_python_call(
         decode_io_object,
         file_json,
-        storage_options=config.SETTINGS["cache_files_storage_options"],
+        storage_options=config.SETTINGS.get().cache_files_storage_options,
         **kwargs,
     )
 
@@ -311,8 +324,9 @@ def dictify_io_object(obj: _UNION_IO_TYPES) -> Dict[str, Any]:
 def register_all() -> None:
     """Register extra encoders if optional dependencies are installed."""
     for type_ in (
-        io.BufferedReader,
-        io.TextIOWrapper,
+        io.RawIOBase,
+        io.BufferedIOBase,
+        io.TextIOBase,
         fsspec.spec.AbstractBufferedFile,
         fsspec.implementations.local.LocalFileOpener,
     ):
