@@ -1,4 +1,5 @@
 """Utilities."""
+
 # Copyright 2019, B-Open Solutions srl.
 # Copyright 2022, European Union.
 #
@@ -16,7 +17,10 @@
 
 import hashlib
 import io
-from typing import Optional, Tuple
+import time
+import warnings
+from types import TracebackType
+from typing import Any, Optional, Tuple, Type
 
 import fsspec
 
@@ -32,14 +36,14 @@ def hexdigestify(text: str) -> str:
 def get_cache_files_fs_dirname() -> Tuple[fsspec.AbstractFileSystem, str]:
     """Return the ``fsspec`` filesystem and directory name where cache files are stored."""
     fs, _, (path,) = fsspec.get_fs_token_paths(
-        config.SETTINGS["cache_files_urlpath"],
-        storage_options=config.SETTINGS["cache_files_storage_options"],
+        config.SETTINGS.get().cache_files_urlpath,
+        storage_options=config.SETTINGS.get().cache_files_storage_options,
     )
     return (fs, path)
 
 
 def copy_buffered_file(
-    f_in: fsspec.spec.AbstractBufferedFile,
+    f_in: Any,
     f_out: fsspec.spec.AbstractBufferedFile,
     buffer_size: Optional[int] = None,
 ) -> None:
@@ -59,4 +63,46 @@ def copy_buffered_file(
         data = f_in.read(buffer_size)
         if not data:
             break
-        f_out.write(data)
+        f_out.write(data if isinstance(data, bytes) else data.encode())
+
+
+class _Locker:
+    def __init__(self, fs: fsspec.AbstractFileSystem, urlpath: str) -> None:
+        self.fs = fs
+        self.urlpath = urlpath
+        self.lockfile = urlpath + ".lock"
+
+    @property
+    def file_exists(self) -> bool:
+        return bool(self.fs.exists(self.urlpath))
+
+    def acquire(self) -> None:
+        self.fs.touch(self.lockfile)
+
+    def release(self) -> None:
+        if self.fs.exists(self.lockfile):
+            self.fs.rm(self.lockfile)
+
+    def wait_until_released(self) -> None:
+        warned = False
+        while self.fs.exists(self.lockfile):
+            if not warned:
+                warnings.warn(
+                    f"can NOT proceed until file is released: {self.lockfile!r}.",
+                    UserWarning,
+                )
+                warned = True
+            time.sleep(1)
+
+    def __enter__(self) -> bool:
+        self.wait_until_released()
+        self.acquire()
+        return self.file_exists
+
+    def __exit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc_val: Optional[BaseException],
+        exc_tb: Optional[TracebackType],
+    ) -> None:
+        self.release()
