@@ -93,17 +93,18 @@ def test_filecache_default() -> None:
 
 @pytest.mark.parametrize("raise_all_encoding_errors", [True, False])
 def test_filecache_default_error(raise_all_encoding_errors: bool) -> None:
+    config.set(raise_all_encoding_errors=raise_all_encoding_errors)
+
     class Dummy:
         pass
 
-    with config.set(raise_all_encoding_errors=raise_all_encoding_errors):
-        if raise_all_encoding_errors:
-            with pytest.raises(AttributeError, match="Can't pickle local object"):
+    if raise_all_encoding_errors:
+        with pytest.raises(AttributeError, match="Can't pickle local object"):
+            encode.filecache_default(Dummy())
+    else:
+        with pytest.warns(UserWarning, match="Can't pickle local object"):
+            with pytest.raises(encode.EncodeError):
                 encode.filecache_default(Dummy())
-        else:
-            with pytest.warns(UserWarning, match="Can't pickle local object"):
-                with pytest.raises(encode.EncodeError):
-                    encode.filecache_default(Dummy())
 
 
 @pytest.mark.parametrize(
